import os
import json

__all__ = ['Jsonl']


class Jsonl:
	@classmethod
	def _traverse(cls, data, f):
		if isinstance(data, dict):
			return {key: cls._traverse(value, f) for key, value in data.items()}
		elif isinstance(data, list):
			return [cls._traverse(item, f) for item in data]
		elif isinstance(data, str):
			return f(data)
		else:
			return data  # Return the data unchanged if it's not a dict, list, or string.

	@classmethod
	def _escape(cls, data):
		return cls._traverse(data, lambda s: s.replace('\n', '\\n'))
	
	@classmethod
	def _unescape(cls, data):
		return cls._traverse(data, lambda s: s.replace('\\n', '\n'))

	@classmethod
	def save(cls, data, name, file_dir, max_file_size=52428800):
		'Writes data to jsonl files ensuring each file does not exceed the specified max_file_size'
		if not isinstance(data, list):
			raise Exception('Data must be a list')
		chunk_num    = 0
		current_size = 0
		file         = None

		def open_new_file():
			nonlocal file, current_size, chunk_num
			if file:
				file.close()
			file_path    = os.path.join(file_dir, f'{name}_{chunk_num}.jsonl')
			file         = open(file_path, 'w')
			current_size = 0
			chunk_num += 1

		open_new_file()

		for item in data:
			jsonl_line = cls.encode(item) + '\n'
			line_size  = len(jsonl_line.encode('utf-8'))
			if max_file_size and current_size + line_size > max_file_size:
				open_new_file()
			file.write(jsonl_line)
			current_size += line_size

		if file:
			file.close()

	@classmethod
	def load_all(cls, name, file_dir):
		'Load all records at once'
		all = []
		files = sorted([os.path.join(file_dir, f) for f in os.listdir(file_dir) if f.startswith(name) and f.endswith('.jsonl')])
		for file_path in files:
			with open(file_path, 'r') as file:
				for line in file:
					all.append(cls.decode(line))
		return all

	@classmethod
	def load(cls, name, file_dir):
		'Generator to read jsonl files in batches'
		files = sorted([os.path.join(file_dir, f) for f in os.listdir(file_dir) if f.startswith(name) and f.endswith('.jsonl')])
		for file_path in files:
			with open(file_path, 'r') as file:
				for line in file:
					yield cls.decode(line)

	@classmethod
	def encode(cls, item):
		return json.dumps(cls._escape(item))

	@classmethod
	def decode(cls, line):
		return cls._unescape(json.loads(line.strip()))


import os
import requests

from .decorators import handle_http_error
from .jsonl      import Jsonl

__all__ = ['Git']


class Git:
	def __init__(self, 
		user,
		repo,
		branch   = 'main',
		host     = 'github.com',
		raw_host = 'raw.githubusercontent.com',
		token    = None,
		encoding = 'utf-8'
	):
		for k, v in locals().items():
			setattr(self, k, v)
	
	def _split_path(self, path):
		dir_name            = os.path.dirname(path)
		file_name           = os.path.basename(path)
		file_name, file_ext = os.path.splitext(file_name) 
		return dir_name, file_name, file_ext

	def _make_raw_url(self, file_path):
		return os.path.join('https://', self.raw_host, self.user, self.repo, self.branch, file_path)

	@handle_http_error
	def _get_response(self, file_path, stream):
		dir_name, file_name, ext_name = self._split_path(file_path)
		url     = self._make_raw_url(file_path)
		headers = {}
		if self.token:
			headers['Authorization'] = f'token {self.token}'
			
		with requests.get(url, headers=headers, stream=stream) as response:
			if response.status_code == 200:
				yield response
			elif response.status_code == 404:
				chunk = 0
				while True:
					file_path = os.path.join(dir_name, f'{file_name}_{chunk}{ext_name}')
					url       = self._make_raw_url(file_path)
					with requests.get(url, headers=headers, stream=stream) as response:
						if response.status_code == 404:
							break
						yield response
					chunk += 1
			else:
				response.raise_for_status()
	
	def load(self, file_path, process_line=None):
		for response in self._get_response(file_path, True):
			for line in response.iter_lines():
				if line and line.strip():
					line = line.decode(response.encoding or self.encoding)
					yield process_line(line) if process_line else line

	def load_all(self, file_path, process_line=None, lines=False) -> list[str] | str:
		if lines:
			return list(self.load(file_path, process_line))
		else:
			s = ''
			for response in self._get_response(file_path, False):
				s += response.content.decode(response.encoding or self.encoding)
			return s

	def load_jsonl(self, file_path, lines=True):
		if lines:
			return self.load(file_path, Jsonl.decode)
		else:
			return self.load_all(file_path)
